// Copyright 2017 - 2018 Modoolar <info@modoolar.com>
// License LGPLv3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html).

odoo.define('scrummer.view.kanban_table', function (require) {
    "use strict";
    const ScrummerData = require('scrummer.data');
    const core = require('scrummer.core');
    const BaseWidgets = require('scrummer.BaseWidgets');
    const TaskWidget = require('scrummer.widget.task').TaskWidget;
    const AgileModals = require('scrummer.widget.modal');
    const storage_service = require('scrummer.storage_service');
    const hash_service = require('scrummer.hash_service');
    require('sortable');
    const pluralize = require('pluralize');
    const DataServiceFactory = require('scrummer.data_service_factory');
    const task_service = DataServiceFactory.get("project.task", false);
    const AgileToast = require('scrummer.toast');
    const dialog = require('scrummer.dialog');
    const mixins = require('scrummer.mixins');

    const web_core = require('web.core');
    const qweb = web_core.qweb;
    const _t = web_core._t;

    const INDEX_NOT_FOUND = -1;

    function getFieldId(record, field) {
        return Array.isArray(record[field]) ? record[field][0] : record[field];
    }

    function getFieldName(record, field) {
        return Array.isArray(record[field]) ? record[field][1] : record[field];
    }

    class SwimlaneDefinition {
        /* eslint-disable no-unused-vars*/
        /**
         *
         * @param {AbstractKanbanTable} kanbanTable - A reference to parent kanbanTable
         * @param {String} field - Name of field that will be used for grouping cards
         * @param {String} text - Will be shown in UI select-option when changing swimlane
         * @param {String} undefinedHeaderName - String shown for items that have undefined value of field
         * @param {String} headerTemplate - Qweb template key that should be rendered for swimlane
         */
        constructor(kanbanTable, field = null, text, undefinedHeaderName = _t("Undefined"), headerTemplate = "scrummer.kanban_table.swimlane.header.simple") {
            this.kanbanTable = kanbanTable;
            this.field = field;
            this.text = text;
            this.undefinedHeaderName = undefinedHeaderName;
            this.headerTemplate = headerTemplate;
            this.swimlaneNoHeaderTemplate = "scrummer.kanban_table.swimlane";
            this.swimlaneTemplate = "scrummer.kanban_table.swimlane.collapsible";
            this.columnTemplate = "scrummer.kanban_table.column";
        }

        /**
         * Returns concrete value or promise which resolves string/number/null that will be used as a group ID for custom grouping of cards in swimlanes.
         * @param {AbstractCard} card
         * @returns {*} - value of field by which cards are grouped
         */
        mapper(card) {
            if (this.field) {
                return getFieldId(card, this.field);
            }
            return null;
        }

        compare(x, y) {
            return x === null || x === false ? 1 : y === null || y === false ? -1 : x - y;
        }

        /**
         * Returns promise which resolves a dictionary that will be passed to qweb.render - headerTemplate
         * If no swimlane is selected, it should return undefined
         * @param {AbstractCard} card
         * @returns {Object}
         */
        header(card) {
            function pluralizedCount() {
                const count = this.recordCount();
                return count + " " + pluralize('issue', count);
            }

            if (this.field === null) {
                return null;
            } else if (!card[this.field]) {
                return {
                    headerData: {name: this.undefinedHeaderName},
                    pluralizedCount
                };
            }
            return {
                headerData: {name: getFieldName(card, this.field)},
                pluralizedCount
            };

        }

        /**
         * Returns concrete value or promise which resolves boolean wether card should be rendered on table or not.
         * @param {AbstractCard} card
         * @returns {Boolean}
         */
        filter(card) {
            return true;
        }

        /**
         * Should be called before any other method.
         * Initializes arrays/maps/etc. of promises used to enable bulk fetching.
         * @returns {Boolean}
         */
        begin() {
            return this;
        }

        /**
         * Should be called when mapper function has been called for all records
         * Purpose of this method is to create a batch request for all records external dependencies.
         * Since base SwimlaneDefinition doesn't return any promises, nothing has to be resolved.
         * References to temporary promises, etc. should be deleted so that GC can clean them up
         * @returns {SwimlaneDefinition}
         */
        resolve() {
            return this;
        }
    }

    class AsyncSwimlaneDefinition extends SwimlaneDefinition {
        mapper(card) {
            if (this.state !== "in_transaction") {
                throw new Error("SwimlaneDefinition mapper called without call to begin method");
            }
        }

        filter(card) {
            if (this.state !== "in_transaction") {
                throw new Error("SwimlaneDefinition filter called without call to begin method");
            }
            return super.filter(card);
        }

        begin() {
            this.state = "in_transaction";
            return this;
        }

        resolve() {
            this.state = "resolved";
            return this;
        }
    }

    class StorySwimlaneDefinition extends AsyncSwimlaneDefinition {
        constructor(kanbanTable) {
            super(kanbanTable, 'story', _t("User Story"), _t("Other issues"), "scrummer.kanban_table.swimlane.header.story");
        }

        begin() {
            super.begin();
            this.parent_ids = [];
            this.parentDeferredMap = {};
            return this;
        }

        _getParentPromise(id) {
            if (!this.parentDeferredMap[id]) {
                this.parentDeferredMap[id] = new $.Deferred();
            }
            this.parent_ids.includes(id) || this.parent_ids.push(id);
            return this.parentDeferredMap[id].promise();
        }

        mapper(task) {
            super.mapper(task);
            if (!task.parent_id) {
                return null;
            }
            const parentPromise = this._getParentPromise(task.parent_id[0]);

            /* eslint-disable-next-line arrow-body-style*/
            return parentPromise.then((parentTask) => {
                return parentTask.is_user_story ? parentTask.id : null;
            });
        }

        filter(task) {
            // Don't render user stories that are already shown as a swimlane
            return Boolean(!(task.is_user_story && task.child_ids.find((t) => this.kanbanTable.data.ids.includes(t))));
        }

        header(task) {
            if (!task.parent_id) {
                const header = super.header(task);
                header.headerData.name = this.undefinedHeaderName;
                header._overrideTemplate = "scrummer.kanban_table.swimlane.header.simple";
                return header;
            }
            return this._getParentPromise(task.parent_id[0]).then((parentTask) => ({

                task: parentTask,
                count() {
                    const count = this.recordCount();
                    return count + " " + pluralize('sub-task', count);
                },
                afterRender: (element) => {
                    element.find(".task-key").click(() => {
                        this.kanbanTable.trigger_up("open_task", {id: parentTask.id});
                    });

                }

            }));
        }

        resolve() {
            super.resolve();
            this.kanbanTable.loadRecords(this.parent_ids).then((records) => {
                for (const record of records) {
                    this.parentDeferredMap[record.id].resolve(record);
                }
            });
            return this;
        }
    }

    const AbstractCard = BaseWidgets.AgileBaseWidget.extend(mixins.MenuItemsMixin, {
        _name: "AbstractCard",
        template: "scrummer.kanban_table.card",
        customCardTitle: null,
        customCardFooter: null,
        init(parent, options) {
            Object.assign(this, options);
            this._super(parent, options);
            this._require_obj("record", ["id"]);
            mixins.MenuItemsMixin.init.call(this);
        },
        renderElement() {
            this._super();
            this.customCardTitle && this.$(".card-title").append(qweb.render(this.customCardTitle, {widget: this}));
            this.customCardFooter && this.$(".footer-content").append(qweb.render(this.customCardFooter, {widget: this}));
        },
        start() {
            mixins.MenuItemsMixin.start.call(this);
            this._is_added_to_DOM.then(() => {
                this.$('.dropdown-button').dropdown();
                this.$('.tooltipped').tooltip();
            });
            return this._super();
        },
        image_url() {
            return false;
        },
        image_tooltip() {
            return false;
        },
        index() {
            return this.$el.index();
        }
    });
    const TaskCard = AbstractCard.extend({
        menuItems: [
            {
                class: "assign-to-me",
                icon: "mdi-account-check",
                text: _t("Assign To Me"),
                callback: '_onAssignToMeClick',
                sequence: 1,
                hidden() {
                    return this.record.user_id && this.record.user_id[0] === ScrummerData.session.uid;
                }
            },
            {
                class: "unassign",
                icon: "mdi-account-minus",
                text: _t("Unassign"),
                callback: '_onUnassignClick',
                sequence: 2,
                hidden() {
                    return !(this.record.user_id && this.record.user_id[0] === ScrummerData.session.uid);
                }
            },
            {
                class: "edit-item",
                icon: "mdi-pencil",
                text: _t("Edit"),
                callback: '_onEditItemClick',
                sequence: 3,
            },
            {
                class: "add-sub-item",
                icon: "mdi-subdirectory-arrow-right",
                text: _t("Add Sub Item"),
                callback: '_onAddSubItemClick',
                sequence: 4,
                hidden() {
                    return !this.task_type.allow_sub_tasks;
                },
            },
            {
                class: "add-link",
                icon: "mdi-link",
                text: _t("Add Link"),
                callback: '_onAddLinkClick',
                sequence: 5,
            },
            {
                class: "work-log",
                icon: "mdi-worker",
                text: _t("Log Work"),
                callback: '_onWorkLogClick',
                sequence: 6,
            },
            {
                class: "add-comment",
                icon: "mdi-comment-account",
                text: _t("Add Comment"),
                callback: '_onAddCommentClick',
                sequence: 7,
            },
            {
                class: "delete",
                icon: "mdi-delete",
                text: _t("Delete"),
                callback: '_onDelete',
                sequence: 8,
            },
        ],
        customCardTitle: "scrummer.kanban_table.card.task.title",
        customCardFooter: "scrummer.kanban_table.card.task.footer",
        init(parent, options) {
            this._super(parent, options);
            this.task = this.record;
        },
        willStart() {
            return this._super().then(() => $.when(ScrummerData.cache.get("current_user").then((user) => ScrummerData.cache.get("team_members", {teamId: user.team_id[0]})).then((members) => {
                this.user = members.find((e) => this.task.user_id && e.id === this.task.user_id[0]);
            }), DataServiceFactory.get("project.task.type2").getRecord(this.task.type_id[0]).then((task_type) => {
                this.task_type = task_type;
            })));
        },
        start() {
            this.$(".task-key").unbind("click");
            this.$(".task-key").click(() => {
                this.trigger_up("open_task", {id: this.task.id});
            });
            return this._super();
        },
        image_url() {
            return this.user ? ScrummerData.getImage("res.users", this.user.id, this.user.write_date) : "/scrummer/static/img/unassigned.png";
        },
        image_tooltip() {
            return this.user ? this.user.name : _t("Unassigned");
        },
        _onWorkLogClick() {
            const modal = new AgileModals.WorkLogModal(this.getParent(), {
                task: this.task,
                userId: ScrummerData.session.uid
            });
            modal.appendTo($("body"));
        },
        _onAddLinkClick() {
            const modal = new AgileModals.LinkItemModal(this.getParent(), {
                task: this.task,
            });
            modal.appendTo($("body"));
        },
        _onAddCommentClick() {
            const modal = new AgileModals.CommentItemModal(this.getParent(), {
                task: this.task,
            });
            modal.appendTo($("body"));
        },
        _onAddSubItemClick() {
            const newItemModal = new AgileModals.NewItemModal(this.getParent(), {
                currentProjectId: this.task.project_id[0],
                parent_id: this.task.id,
            });
            newItemModal.appendTo($("body"));
        },
        _onEditItemClick() {
            const newItemModal = new AgileModals.NewItemModal(this.getParent(), {
                currentProjectId: this.task.project_id[0],
                edit: this.task,
            });
            newItemModal.appendTo($("body"));
        },
        _onAssignToMeClick() {
            this.task.user_id = ScrummerData.session.uid;
            // data.cache.get("current_user").then(user => {
            //     this.user = user;
            //     this.renderElement();
            //     this.start();
            // });
        },
        _onUnassignClick() {
            this.task.user_id = false;
        },
        _onDelete() {
            dialog.confirm(_t("Delete task"), _t("Are you sure you want to delete this task?"), _t("yes")).done(() => {
                this.task.unlink();
            });
        },

    });

    const AbstractKanbanTable = BaseWidgets.AgileBaseWidget.extend({
        _name: "KanbanTable",
        template: "scrummer.kanban_table",
        kanbanTableOptionsID: null,
        init(parent, options) {
            Object.assign(this, options);
            this._super(parent, options);
            this.swimlanes = this.initSwimlaneDefinitions();
            this._require_prop("swimlanes", "This property specifies fields used to group cards in swimlanes. " +
                "Fields are stored in ES6 Map." +
                "Key is field name, value is instance of SwimlaneDefinition ES6 class");
            this._require_prop("dataService", "An instance of implementation of DataService from scrummer.DataService module.");
            this._require_prop("stageField", "This property tells table what is the field of card model that workflow is wrapped around.");
            this._require_prop("kanbanTableOptionsID", "This property is used to identify key of options object in local storage");
            /**
             * This method should be overridden in concrete implementation
             * Here we have demo structure that shows what loadData should resolve
             {
                board: {
                    name: "Development",
                    columns: {
                        "1": {order: 1, id: 1, name: "To Do"},
                        "2": {order: 2, id: 2, name: "In Progress"},
                        "3": {order: 3, id: 3, name: "Done"},
                    },
                    // state_id is link to project.workflow.state
                    status: {
                        "1": {state_id: 72, "column_id": 1, "id": 1, "order": 2},
                        "2": {state_id: 66, "column_id": 1, "id": 2, "order": 1},
                        "3": {state_id: 67, "column_id": 2, "id": 3, "order": 0},
                        "4": {state_id: 68, "column_id": 3, "id": 4, "order": 0}
                    },
                },
                workflow: {
                    workflows: {
                        1: {id: 1, name: "Workflow 1", description: "Workflow 1 description"},
                        2: {id: 2, name: "Workflow 2", description: "Workflow 2 description"},
                        3: {id: 3, name: "Workflow 3", description: "Workflow 3 description"},
                    }
                    states: {
                        66: {
                            id: 66,
                            stage_id: 4, // This will has to be defined when creating KanbanTable (stageField)
                            global_in: true,
                            global_out: false,
                            name: "To Do",
                            type: "todo",
                            workflow_id: 1,
                            in_transitions: [],
                            out_transitions: [91],
                        },
                        67: {
                            id: 67,
                            stage_id: 14,
                            global_in: false,
                            global_out: false,
                            name: "In Progress",
                            type: "in_progress",
                            workflow_id: 1,
                            in_transitions: [91, 97],
                            out_transitions: [92, 95]
                        },
                        ...
                    }
                    transitions: {
                        1: {
                            id: 1,
                            name: "Start progress",
                            description: "Begin working on task",
                            src: 66,
                            dst: 67,
                            workflow_id: 1,
                            user_confirmation: true,
                        },
                        ...
                    }
                }
                ids: [1,2,3,4,5]
             }
             */
            this._require_obj("data", ['board', 'workflow', 'ids'], "Check code comments for propper format of data");
            this._require_obj("Card", ["Card"], "Card object with Card property must point to concrete implementation of AbstractCard class.");

            this.records = [];

            //Keeping track of card widgets
            this.cardWidgetsMap = new Map();

            // Load current board options from storage service
            this.kanbanOptions = storage_service.get("kto" + this.kanbanTableOptionsID) || {};
            this.swimlane = this.swimlane || this.kanbanOptions.swimlane || null;

        },
        /**
         * Override this method in order to append more swimlane options
         * @returns {Map}
         */
        initSwimlaneDefinitions() {
            return new Map([
                [null, new SwimlaneDefinition(this, null, _t("No swimlanes"))],
            ]);
        },
        willStart() {
            // Only user result is needed to be assigned to widget.
            return $.when(ScrummerData.cache.get("current_user"), this._super.apply(arguments), this.loadRecords(this.data.ids)).then((user) => {
                    this.current_user = user;
                    this.loadData();
                    return this.prepareSwimlaneData();
                }
            );
        },
        /**
         * Wraps dataService.getRecords and saves records in cache for later synchronous use.
         * @param {Number[]} ids
         * @returns {jQuery.Promise}
         */
        loadRecords(ids) {
            return this.dataService.getRecords(ids).then((records) => {
                this.records.push.apply(this.records, records);
                return records;
            });
        },
        /**
         * Wraps dataService.getRecord and saves record in cache for later synchronous use.
         * @param {Number} id
         * @returns {jQuery.Promise}
         */
        loadRecord(id) {
            return this.dataService.getRecord(id).then((record) => {
                this.records.push(record);
                return record;
            });
        },
        loadData() {
            // Create columns array and sort it by column order
            const columns_map = this.data.board.columns;
            this.data.sorted_columns = Object.keys(columns_map).map((key) => columns_map[key]).sort((a, b) => a.order - b.order);
            this.states = {};
            this.data.board.stateToStatus = {};
            this.data.workflow.global_states = {
                in: [],
                out: []
            };
            // Assign every state to a column in sorted_columns, map state to status and save global states
            for (const status_id in this.data.board.status) {
                const status = this.data.board.status[status_id];
                const state = this.data.workflow.states[status.state_id];
                const column = this.data.sorted_columns.find((col) => status.column_id === col.id);

                state.global_in && this.data.workflow.global_states.in.push(state);
                state.global_out && this.data.workflow.global_states.out.push(state);

                this.data.board.stateToStatus[state.id] = status;

                // Push state to column.states and create array if it didn't exist before.
                (column.status = column.status || []).push(status);
                this.states[state.id] = state;
            }
        },
        renderElement() {
            this._super();
            this.$(".column-headers").append(this.renderHeaders());
            this.renderSwimlanes(this.$(".swimlanes"));
        },
        prepareSwimlaneData() {
            const swimlaneDefinition = this.swimlanes.get(this.swimlane).begin();
            this.data.swimlanes = new Map();

            const promises = [];

            function recordCount() {
                return this.records.length;
            }

            for (const record_id of this.data.ids) {
                const record = this.records.find((r) => r.id === record_id);
                promises.push($.when(swimlaneDefinition.mapper(record), swimlaneDefinition.filter(record)).then((group, passIt) => {
                    if (passIt) {
                        if (this.data.swimlanes.has(group)) {
                            this.data.swimlanes.get(group).records.push(record);
                        } else {
                            const swimlaneData = {records: [record]};
                            this.data.swimlanes.set(group, swimlaneData);
                            return $.when(swimlaneDefinition.header(record)).then((header) => {
                                if (header) {
                                    header.recordCount = recordCount.bind(swimlaneData);
                                }
                                swimlaneData.header = header;
                            });
                        }
                    }
                }));
            }
            swimlaneDefinition.resolve();
            return $.when(...promises);
        },
        getActiveSwimlane() {
            return this.swimlanes.get(this.swimlane);
        },
        renderHeaders() {
            const result = [];
            for (const column of this.data.sorted_columns) {
                result.push(`<li class="column" data-id="${column.id}">${column.name}</li>`);
            }
            return result;
        },
        renderSwimlanes(placeholder) {
            const result = [];
            if (this.data.swimlanes.size < 1) {
                return;
            } else if (this.data.swimlanes.size === 1 && [...this.data.swimlanes.keys()][0] === null) {
                this.data.swimlanes.forEach((data, id) => {
                    result.push(this.renderSwimlane(data, id, false, placeholder));
                });
            } else {
                const swimlaneDefinition = this.swimlanes.get(this.swimlane);

                // let sortedSwimlanes = [...this.data.swimlanes.keys()].sort((a, b) => !a ? 1 : -1);
                const sortedSwimlanes = [...this.data.swimlanes.keys()].sort(swimlaneDefinition.compare);
                sortedSwimlanes.forEach((id) => {
                    const data = this.data.swimlanes.get(id);
                    result.push(this.renderSwimlane(data, id, true, placeholder));
                });
            }
            return result;
        },
        renderSwimlane(data, id, renderHeader = true, placeholder) {
            let node = null;
            const swimlaneDefinition = this.swimlanes.get(this.swimlane);

            if (renderHeader) {
                // If swimlane has header, render collapsibles
                node = $(qweb.render(swimlaneDefinition.swimlaneTemplate).trim());
                const headerTemplate = data.header._overrideTemplate || swimlaneDefinition.headerTemplate;
                const header = $(qweb.render(headerTemplate, data.header).trim());
                node.find(".collapsible-header").append(header);
                if (typeof data.header.afterRender === "function") {
                    data.header.afterRender(header);
                }
            } else {
                // No swimlane
                node = $(qweb.render(swimlaneDefinition.swimlaneNoHeaderTemplate).trim());
            }

            node.attr("data-swimlane-id", id);

            placeholder.append(node);

            // prepare records for columns and render them
            const recordsByStageFieldId = new Map();
            data.records.forEach((record) => {
                const recordStageFieldId = this.getStageFieldId(record);
                recordsByStageFieldId.has(recordStageFieldId)
                    ? recordsByStageFieldId.get(recordStageFieldId).push(record)
                    : recordsByStageFieldId.set(recordStageFieldId, [record]);
            });


            for (const column of this.data.sorted_columns) {
                const column_cards = [];
                recordsByStageFieldId.forEach((records, stageFieldId) => {
                    // Check if card belongs to column by checking if some status has workflow.state that wraps around records stageFieldId
                    if (column.status && column.status.some((e) => this.data.workflow.states[e.state_id].stage_id === stageFieldId)) {
                        column_cards.push(...records);
                    }
                });
                this.renderColumn(column, column_cards, node.find(".columns"));
            }
            return node;
        },
        renderColumn(column, records, placeholder) {
            const swimlaneDefinition = this.swimlanes.get(this.swimlane);
            const node = $(qweb.render(swimlaneDefinition.columnTemplate, column));
            placeholder.append(node);
            const sorted_records = records.sort((a, b) => a.agile_order - b.agile_order);
            for (const record of sorted_records) {
                const cardWidget = this.renderCard(record);
                cardWidget.appendTo(node).then(() => {
                    cardWidget.start();
                });
            }

            const tableThis = this;
            node.sortable({
                start: function (evt, data) {
                    const cardNode = $(data.helper);
                    const swimlaneId = cardNode.getDataFromAncestor("swimlane-id", ".swimlane");
                    const columnId = cardNode.getDataFromAncestor("column-id");
                    const cardId = cardNode.getDataFromAncestor("card-id");

                    this.swimlaneNode = swimlaneId === null
                        ? tableThis.$(`.swimlane:not([data-swimlane-id])`)
                        : tableThis.$(`.swimlane[data-swimlane-id=${swimlaneId}]`);

                    let preventStopEvent = this.preventStopEvent;
                    // Generate overlays over all other columns except the source column
                    for (let columnNode of $.makeArray(this.swimlaneNode.find(`.column:not([data-column-id=${columnId}])`))) {
                        columnNode = $(columnNode);
                        const columnNodeId = columnNode.data("column-id");
                        const columnOverlay = tableThis.generateOverlay(columnNodeId, cardId);

                        // Attach event on drop on each of overlay states
                        columnOverlay.find(".state").on("drop", (event, ui) => {
                            const droppedCardNode = $(ui.helper);
                            const droppedCardId = droppedCardNode.data("card-id");
                            const targetDataset = event.target.dataset;
                            const newState = tableThis.data.workflow.states[targetDataset.stateId];
                            const newStatus = tableThis.data.board.stateToStatus[newState.id];
                            const isTransitionGlobal = targetDataset.transitionId === "global";
                            const transition = tableThis.data.workflow.transitions[targetDataset.transitionId];

                            // get cardWidget from table
                            const cardWidget = tableThis.cardWidgetsMap.get(droppedCardId);
                            const stage_id = parseInt(newState.stage_id, 10);

                            if (!isTransitionGlobal && transition.user_confirmation) {
                                tableThis.openStageChangeModal(newState.stage_id, cardWidget);
                            } else {
                                cardWidget.record[tableThis.stageField] = stage_id;
                            }

                            // Set property of sortable so that sorting within column is not mixed with changing column.
                            preventStopEvent = true;
                        });
                        columnOverlay.prependTo(columnNode);
                    }
                },
                stop: function (event, ui) {
                    this.swimlaneNode.find(".column-overlay").remove();

                    if (this.preventStopEvent) {
                        delete this.preventStopEvent;
                        return;
                    }
                }
            });
            return node;
        },
        renderCard(record) {
            if (this.cardWidgetsMap.has(record.id)) {
                return this.cardWidgetsMap.get(record.id);
            }
            const cardWidget = new this.Card.Card(this, {record});
            this.cardWidgetsMap.set(record.id, cardWidget);
            return cardWidget;

        },
        getStageFieldId(record) {
            return getFieldId(record, this.stageField);
        },
        getStageFieldName(record) {
            return getFieldName(record, this.stageField);
        },
        getWorkflowState(stageFieldId, workflow_id) {
            for (const state_id in this.data.workflow.states) {
                const state = this.data.workflow.states[state_id];
                if (state.workflow_id === workflow_id && state[this.stageField] === stageFieldId) {
                    return state;
                }
            }
        },
        getColumnFromStageField(stageFieldId) {
            for (const column of this.data.sorted_columns) {
                for (const status of column.status) {
                    if (this.data.workflow.states[status.state_id].stage_id === stageFieldId) {
                        return column;
                    }
                }
            }
        },
        generateOverlay: function (columnId, recordId) {
            const record = this.records.find((r) => r.id === recordId);
            const stageFieldId = this.getStageFieldId(record);

            const currentState = this.getWorkflowState(stageFieldId, record.workflow_id[0]);
            const availableTransitions = this.getAvailableTransitions(currentState, this.data.workflow, record);

            const overlay = $(`<div class="column-overlay" data-column-id="${columnId}"></div>`);

            if (this.data.board.columns[columnId].status) {
                for (const status of this.data.board.columns[columnId].status) {
                    const state = this.data.workflow.states[status.state_id];
                    const transition = availableTransitions.find((t) => t.dst === state.id);
                    if (transition) {
                        const stateNode = $(`<div class="state agile-main-color lighten-5" data-state-id="${state.id}" data-transition-id="${transition.id}">${state.name}</div>`);
                        stateNode.droppable({
                            classes: {"ui-droppable-hover": "hover"},
                        });
                        stateNode.appendTo(overlay);
                    }
                }
            }
            return overlay;
        },
        getAvailableTransitions(currentState, workflow) {
            // Duplicate transitions won't be rendered multiple times, so don't worry
            // This case can happen, if there is in/out transition and state is global in/our
            const transitions = currentState.out_transitions.map((t_id) => workflow.transitions[t_id]);
            if (currentState.global_out) {
                for (const state_id in workflow.states) {
                    const state = workflow.states[state_id];
                    if (state.id !== currentState.id && state.workflow_id === currentState.workflow_id) {
                        transitions.push(this.generateFakeTransition(workflow, currentState, state));
                    }
                }
            } else {
                workflow.global_states.in.filter((state) => state.id !== currentState.id && state.workflow_id === currentState.workflow_id).forEach((state) => {
                    transitions.push(this.generateFakeTransition(workflow, currentState, state));
                });
            }
            return transitions;
        },
        generateFakeTransition(workflow, src, dst) {
            return {
                description: "",
                dst: dst.id,
                id: "global",
                name: dst.name,
                src: src.id,
                user_confirmation: false,
                workflow_id: workflow.id
            };
        },
        openStageChangeModal(newStageId, cardWidget, confirmedCallback) {
            const state = this.getWorkflowState(newStageId, cardWidget.record.workflow_id[0]);
            const modal = new AgileModals.TaskStageConfirmationModal(this, {
                taskId: cardWidget.record.id,
                stageId: newStageId,
                stageName: state.name,
                userName: cardWidget.record.user_id ? cardWidget.record.user_id[1] : _t("Unassigned"),
                afterHook: (result) => {
                    typeof confirmedCallback === "function" && confirmedCallback(result);
                }
            });
            modal.appendTo($("body"));
        },
        getCard(id) {
            return this.cardWidgetsMap.get(id);
        },
        addCard(id, syncerMeta, highlight = true) {
            return this._doAddCard(id, syncerMeta, highlight, INDEX_NOT_FOUND);
        },

        insertCardAt(id, index, syncerMeta, highlight = true) {
            return this._doAddCard(id, syncerMeta, highlight, index);
        },
        shouldCardBeAdded(record) {
            return true;
        },
        _doAddCard(id, syncerMeta, highlight, index) {
            return this.dataService.getRecord(id).then((record) => {
                if (record && this.shouldCardBeAdded(record)) {
                    const swimlaneDefinition = this.swimlanes.get(this.swimlane).begin();
                    const def = $.when(swimlaneDefinition.mapper(record), swimlaneDefinition.filter(record)).then((group, passIt) => {
                        if (passIt) {
                            if (this.data.swimlanes.has(group)) {
                                this.data.swimlanes.get(group).records.push(record);
                                const column = this.getColumnFromStageField(this.getStageFieldId(record));
                                if (!column) {
                                    return;
                                }
                                const node = group === null
                                    ? this.$(`.swimlane:not([data-swimlane-id]) ${" "} div[data-column-id="${column.id}"]`)
                                    : this.$(`.swimlane[data-swimlane-id=${group}] ${" "} div[data-column-id="${column.id}"]`);
                                const cardWidget = this.renderCard(record);

                                if (index > INDEX_NOT_FOUND) {
                                    cardWidget.insertAt(node, index).then(() => {
                                        cardWidget.start();
                                    });

                                } else {
                                    cardWidget.appendTo(node).then(() => {
                                        cardWidget.start();
                                    });

                                }
                                const swimlaneNode = group === null
                                    ? this.$('.swimlane:not([data-swimlane-id])')
                                    : this.$(`.swimlane[data-swimlane-id=${group}]`);
                                const count = this.data.swimlanes.get(group).records.length;
                                swimlaneNode.find(".swimlane-header .count").text(count + " " + pluralize('issue', count));

                                if (syncerMeta.indirect === false) {
                                    cardWidget._is_added_to_DOM.then(() => {
                                        node.scrollToElement(cardWidget.$el);
                                        highlight && cardWidget.$el.highlight();
                                    });
                                }
                            } else {
                                // If destination swimlane doesn't exist then render it with new card and append in view
                                const swimlaneData = {records: [record]};
                                this.data.swimlanes.set(group, swimlaneData);

                                const recordCount = function () {
                                    return this.records.length;
                                };
                                return $.when(swimlaneDefinition.header(record)).then((header) => {
                                    if (header) {
                                        header.recordCount = recordCount.bind(swimlaneData);
                                    }
                                    swimlaneData.header = header;
                                    // If last swimlane is undefined, then insert new swimlane before it, otherwize insert swimlane after it.
                                    const lastSwimlane = this.$(".swimlane").last();
                                    const newSwimlaneNode = this.renderSwimlane(swimlaneData, group, Boolean(header), lastSwimlane);
                                    if (this.$(".swimlane").length === 0) {
                                        this.$(".swimlanes").append(newSwimlaneNode);
                                    } else if (lastSwimlane.data("swimlane-id")) {

                                        const sortedSwimlanes = [...this.data.swimlanes.keys()].sort(swimlaneDefinition.compare);
                                        const i = sortedSwimlanes.indexOf(group);

                                        this.$(".swimlanes").insertAt(newSwimlaneNode, i);
                                    } else {
                                        newSwimlaneNode.insertBefore(lastSwimlane);
                                    }
                                    // Enable collapsing in new swimlane
                                    newSwimlaneNode.find('.collapsible').collapsible();
                                });
                            }
                            if (syncerMeta) {
                                if (syncerMeta.user_id.id !== ScrummerData.session.uid && syncerMeta.indirect === false) {
                                    AgileToast.toastTask(syncerMeta.user_id, record, syncerMeta.method);
                                }
                            }
                        }
                    });
                    swimlaneDefinition.resolve();
                    return def;
                }
            });

        },
        removeCard(id, removeFromCache = false, syncerMeta) {
            const cardWidget = this.cardWidgetsMap.get(id);
            if (!cardWidget) {
                return false;
            }
            if (removeFromCache) {
                const swimlaneDefinition = this.swimlanes.get(this.swimlane).begin();
                const previousMapped = cardWidget.record._previous ? swimlaneDefinition.mapper(cardWidget.record._previous) : null;
                $.when(previousMapped, swimlaneDefinition.mapper(cardWidget.record), swimlaneDefinition.filter(cardWidget.record)).then((previousGroup, group, passIt) => {
                    if (previousGroup !== null) {
                        const swimlaneData = this.data.swimlanes.get(previousGroup);
                        swimlaneData.records = swimlaneData.records.filter((e) => e.id !== cardWidget.record.id);
                        const swimlaneNode = previousGroup === null
                            ? this.$('.swimlane:not([data-swimlane-id])')
                            : this.$(`.swimlane[data-swimlane-id=${previousGroup}]`);
                        if (swimlaneData.records.length === 0) {
                            this.data.swimlanes.delete(previousGroup);
                            swimlaneNode.remove();
                        } else {
                            const count = swimlaneData.records.length;
                            swimlaneNode.find(".swimlane-header .count").text(count + " " + pluralize('issue', count));
                        }
                    }
                });

                cardWidget.destroy();
                this.cardWidgetsMap.delete(id);
            } else {
                cardWidget.$el.detach();
            }
            if (syncerMeta) {
                if (syncerMeta.user_id.id !== ScrummerData.session.uid && syncerMeta.indirect === false) {
                    AgileToast.toastTask(syncerMeta.user_id, syncerMeta.data, syncerMeta.method);
                }
            }
        },
        /**
         * This method should return [project.workflow.state].id.
         * Usually from <model>[this.stageField] and workflow,
         * but if there is related field on model, it can be returned also.
         *
         * @param {Number} id - ID of record under the card.
         */
        getCardState(id) {
            throw new Error("Not implemented");
        },
        resolveCardIndex(card, delta) {
            return INDEX_NOT_FOUND;
        },
        setSwimlane(name, store = true) {
            this.kanbanOptions.swimlane = name;
            store && storage_service.set("kto" + this.kanbanTableOptionsID, this.kanbanOptions);
            this.prepareSwimlaneData().then(() => {
                this.$(".swimlanes").empty();
                this.renderSwimlanes(this.$(".swimlanes"));
                this.$('.collapsible').collapsible();
            });
        },
        addedToDOM() {
            this.$('.collapsible').collapsible();
        }
    });
    const TaskTable = AbstractKanbanTable.extend({
        Card: {Card: TaskCard},
        dataService: task_service,
        stageField: "stage_id",
        init(parent, options) {
            this._super(parent, options);
            if (!Array.isArray(this.data.board.task_types)) {
                throw new Error("TaskTable requires board to define task_type as Array");
            }
        },
        initSwimlaneDefinitions() {
            const defs = this._super();
            defs.set('user_id', new SwimlaneDefinition(this, 'user_id', _t("Assignee"), _t("Unassigned")));
            defs.set('priority_id', new SwimlaneDefinition(this, 'priority_id', _t("Priority"), _t("Without priority")));
            defs.set('project_id', new SwimlaneDefinition(this, 'project_id', _t("Project"), _t("Without project")));
            defs.set('epic_id', new SwimlaneDefinition(this, 'epic_id', _t("Epic"), _t("Without epic")));
            defs.set('story', new StorySwimlaneDefinition(this));
            return defs;
        },
        loadData() {
            this._super();
            this.board_project_ids = Object.keys(this.data.board.projects).map((k) => parseInt(k, 10));
        },
        getCardState(id) {
            return this.cardWidgetsMap.get(id).task.wkf_state_id;
        },
        shouldCardBeAdded(task) {

            // Filter tasks by project filters
            const team_project_ids = this.current_user.team_ids[this.current_user.team_id[0]].project_ids;
            const hash_project_id = parseInt(hash_service.get("project"), 10);
            const task_project_id = task.project_id[0];

            if (hash_project_id && task_project_id !== hash_project_id) {
                return false;
            }
            if (!team_project_ids.includes(task_project_id)) {
                return false;
            }
            if (!this.board_project_ids.includes(task_project_id)) {
                return false;
            }

            // Filter task by task_types filter
            if (this.data.board.task_types.length && !this.data.board.task_types.includes(task.type_id[0])) {
                return false;
            }

            // todo Map task (stage_id, workflow_id) to Wkf State and then check if the state is mapped to any of the columns in the board
            // todo Check if stage_id is in this.data.workflow.states (but create stageToState map when loading)

            return true;
        },
        resolveCardIndex(id, delta) {
            const card = this.cardWidgetsMap.get(id);
            if (!card) {
                return INDEX_NOT_FOUND;
            }

            const swimlane = this.getActiveSwimlane();
            const swimlane_field = swimlane.field === null ? false
                : swimlane.field === 'story' ? 'parent_id' : swimlane.field;

            const isUpdated = function (fieldName) {
                if (!(fieldName in delta)) {
                    return false;
                }
                const left = getFieldId(delta[fieldName]);
                const right = fieldName in card.record._previous ? getFieldId(card.record._previous[fieldName]) : false;
                return left !== right;
            };

            const isUpdatedStageField = function () {
                return isUpdated(this.stageField);
            }.bind(this);


            // When there is no swimlane we need to see if the stage is updated
            if (swimlane_field === null) {
                return isUpdatedStageField();
            }

            // Check if the current swimlane field is updated
            if (isUpdated(swimlane.field)) {
                return INDEX_NOT_FOUND;
            }

            if (isUpdatedStageField()) {
                return INDEX_NOT_FOUND;
            }

            return card.index();
        },
    });

    const AbstractKanbanTableView = BaseWidgets.AgileViewWidget.extend({
        _name: "KanbanTableView",
        template: "scrummer.view.kanban_table",
        emptyTitle: _t("No data in kanban table"),
        emptyTemplate: "scrummer.view.kanban_table.empty",
        init(parent, options) {
            Object.assign(this, options);
            this._super(parent, options);
            // Because of core.Class implementation wraps functions, we  will have to wrap KanbanTable class in an object.
            this._require_obj("KanbanTable", ["KanbanTable"], "KanbanTable object with KanbanTable property must point to concrete implementation of AbstractKanbanTable class.");
        },
        openSettings() {
            // Set form fields.
            if (this.kanbanTable.swimlane) {
                this.$("#modal-settings select").val(this.kanbanTable.swimlane).material_select();
            }
            const modalNode = this.$("#modal-settings");
            modalNode.materialModal("open");
        },
        destroy() {
            this._super();
            // Not sure if this is needed anymore
            // TODO: Check why is this here...
            if ($(".lean-overlay").size()) {
                $(".lean-overlay").remove();
            }
        },
        isEmpty() {
            throw new Error("Not implemented");
        },
        getTitle() {
            console.warn("This should be overridden in concrete implementation");
            return _t("Kanban table");
        },
        willStart() {
            return $.when(this._super(), ScrummerData.cache.get("current_user").then((user) => {
                this.user = user;
            }));
        },
        renderElement() {
            // Handling case when there is no data, and empty template should be rendered
            if (this.isEmpty()) {
                this.setTitle(this.emptyTitle);
            } else {
                this.title = this.getTitle();
            }
            this._super();
            if (!this.isEmpty()) {
                this.kanbanTable = new this.KanbanTable.KanbanTable(this, this.generateKanbanTableOptions());

                // Render settings modal
                this.$("#kanban-table-view").append(this.renderSettingsModal());
                this.kanbanTable.appendTo(this.$("#kanban-table-view"));
                this.setTitle(this.getTitle());

                // Set action menu
                this.trigger_up("init_action_menu", this.getActionMenu());
            }
            //this.$el.append(JSON.stringify(this.data));
        },
        generateKanbanTableOptions() {
            return {data: this.data};
        },
        renderSettingsModal() {
            const swimlanes = [];
            for (const [key, obj] of this.kanbanTable.swimlanes.entries()) {
                swimlanes.push({value: key, text: obj.text});
            }
            return qweb.render("scrummer.view.kanban_table.settings_modal", {swimlanes});
        },
        _getSwimlaneOptions() {
            return this.kanbanTable.swimlanes;
        },
        getActionMenu() {
            return {
                items: [
                    {icon: "settings", action: this.openSettings.bind(this)}
                ]
            };
        },
        start() {
            this._is_added_to_DOM.then(() => {
                this.$('select').material_select();
                //this.$('.section').perfectScrollbar({suppressScrollY: true});
            });

            const modalNode = this.$("#modal-settings");
            // On save button click check if form has been modified, and if so update options, and store them to storage_service
            this.$("#modal-settings .modal-save").click((evt) => {
                if (this.kanbanTable.swimlane !== this.$("#modal-settings select :selected").attr("value")) {
                    this.kanbanTable.swimlane = this.$("#modal-settings select :selected").attr("value");
                    this.kanbanTable.setSwimlane(this.kanbanTable.swimlane);
                }
                modalNode.materialModal("close");
            });
            modalNode.materialModal();
            return this._super();
        }
    });
    const TaskKanbanTableView = AbstractKanbanTableView.extend({
        custom_events: Object.assign({}, AbstractKanbanTableView.prototype.custom_events, {
            open_task: "_onOpenTask",
        }),
        _onOpenTask: function (evt) {
            this.trigger_up("open_right_side", {
                WidgetClass: TaskWidget,
                options: {id: evt.data.id, isQuickDetailView: true}
            });
        },
        addedToDOM() {
            core.bus.on("project.task:write", this, this._onProjectTaskWrite);
            core.bus.on("project.task:create", this, this._onProjectTaskCreate);
            core.bus.on("project.task:unlink", this, this._onProjectTaskUnlink);
            return this._super();
        },
        _onProjectTaskWrite(id, delta, payload, record) {
            if (!this.kanbanTable) {
                return;
            }

            const editPromise = record && record._edit("check") ? record._edit() : $.when();
            editPromise.then(() => {
                const cardIndex = this.kanbanTable.resolveCardIndex(id, delta);
                this.kanbanTable.removeCard(id, true);
                return this.kanbanTable.insertCardAt(id, cardIndex, payload).then(() => {
                    if (this.rightSideWidget && this.rightSideWidget.id === id) {
                        // Since trigger_up wraps event arguments in data object, here I mimic that behaviour.
                        this.trigger("open_right_side", {
                            data: {
                                WidgetClass: TaskWidget,
                                options: {id, isQuickDetailView: true}
                            }
                        });
                    }
                });
            });
        },
        _onProjectTaskCreate(id, vals, payload) {
            if (!this.kanbanTable) {
                return;
            }

            this.kanbanTable.addCard(id, payload);
        },
        _onProjectTaskUnlink(id, payload) {
            if (!this.kanbanTable) {
                return;
            }

            this.kanbanTable.removeCard(id, true, payload);
            if (this.rightSideWidget && this.rightSideWidget.id === id) {
                this.rightSideWidget.destroy(true);
                delete this.rightSideWidget;
            }
        },
    });


    return {
        AbstractKanbanTableView,
        AbstractKanbanTable,
        AbstractCard,
        TaskCard,
        TaskKanbanTableView,
        TaskTable,
        SwimlaneDefinition,
        AsyncSwimlaneDefinition,
        StorySwimlaneDefinition,
    };
});
